// -*- C++ -*-

#pragma once

#include "gpu_types.h"
#include "loops.h"
#include "buffer.h"

/// \addtogroup thread_communication
/// \{
/// These functions are provided to exchange data between threads in the workgroup.


namespace goopax
{
  namespace impl
  {
    namespace interface
    {
      extern "C"
      {
          GOOPAX_DEPRECATED GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_shuffle(gpu_writer_base_impl**, Uint, const gpu_writer_base_impl*, Uint) noexcept;
	  GOOPAX_DLL_EXPORT(3.1.1) errno_type goopax_E_shuffle2(void*, const gpu_writer_base_impl*, Uint, const gpu_struct::mapframe_base&) noexcept;
          GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_ballot(gpu_writer_base_impl**, Uint, const gpu_writer_base_impl*) noexcept;
          GOOPAX_DLL_EXPORT(3.0.0) Bool goopax_support_vote(Uint, gpu_writer_base_impl*) noexcept;
	GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_work_group_all2(gpu_writer_base_impl*, gpu_writer_base_impl*) noexcept;
	GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_work_group_any2(gpu_writer_base_impl*, gpu_writer_base_impl*) noexcept;
	GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_work_group_reduce_add2(gpu_writer_base_impl*, gpu_writer_base_impl*) noexcept;
	GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_work_group_reduce_min2(gpu_writer_base_impl*, gpu_writer_base_impl*) noexcept;
	GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_work_group_reduce_max2(gpu_writer_base_impl*, gpu_writer_base_impl*) noexcept;
	GOOPAX_DEPRECATED GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_work_group_broadcast2(gpu_writer_base_impl*, gpu_writer_base_impl*, gpu_writer_base_impl*) noexcept;
	GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_work_group_scan_exclusive_add2(gpu_writer_base_impl*, gpu_writer_base_impl*) noexcept;
	GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_work_group_scan_inclusive_add2(gpu_writer_base_impl*, gpu_writer_base_impl*) noexcept;
	GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_work_group_scan_exclusive_min2(gpu_writer_base_impl*, gpu_writer_base_impl*) noexcept;
	GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_work_group_scan_inclusive_min2(gpu_writer_base_impl*, gpu_writer_base_impl*) noexcept;
	GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_work_group_scan_exclusive_max2(gpu_writer_base_impl*, gpu_writer_base_impl*) noexcept;
	GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_work_group_scan_inclusive_max2(gpu_writer_base_impl*, gpu_writer_base_impl*) noexcept;
      }
    }

      
    /*      /// \brief Exchanges values between threads in the workgroup.
      /// \ingroup thread_communication
      /// \param value value to share to other threads
      /// \param source_thread the value is read from this thread within the sub-group. source must have a value between 0 and width-1.
      /// \param width sub-group width in which to exchange the values. width must be a power of 2 and/or equal to local_size(). width must not exceed local_size(). For nvidia GPUs using the CUDA backend, it is suggested that width has a value of 32 or less, in order to take advantage of CUDA's warp shuffle functions.
      /// \return value read from thread 'source'
      template <typename T> gpu_type<T> shuffle(gpu_type<T> value, gpu_uint source_thread, Tuint width)
      {
	  array<gpu_type<T>, 1> tmp = {{value}};
	  return shuffle(tmp, source_thread, width)[0];
      }
    */
    
      /// \brief Exchanges values between threads in the workgroup.
      /// \ingroup thread_communication
      /// \brief value T goopax struct data type
      /// \param value value to share to other threads
      /// \param source_thread the value is read from this thread within the sub-group. source_thread must have a value between 0 and width-1.
      /// \param width sub-group width in which to exchange the values. width must be a power of 2 and/or equal to local_size(). width must not exceed local_size(). For nvidia GPUs using the CUDA backend, it is suggested that width has a value of 32 or less, in order to take advantage of CUDA's warp shuffle functions.
      /// \return value read from thread 'source_thread'
    template <typename T, typename=typename enable_if<is_same<T, typename gettype<T>::gpu>::value>::type>
    T shuffle(T value, gpu_uint source_thread, Tuint width)
      {
          static_assert(is_same<T, typename gettype<T>::gpu>::value, "T needs to be a goopax struct of gpu type.");
          const gpu_struct::mapframe_base& gpu_map = mapframe<typename goopax_struct_changetype<T, xtype_gpu>::type>::get_static_mf();

	  check_errno(interface::goopax_E_shuffle2(&value, source_thread.impl.get(), width, gpu_map));
	  return value;
      }
      
      /// \brief collect booleans from threads of the workgroup into integer bitset
      /// \ingroup thread_communication
      /// \param value boolean to contribute from the current thread
      /// \param width sub-group width for the ballot operation. width must be a power of 2 and must not exceed local_size(). For nvidia GPUs using the CUDA backend, it is suggested that width has a value of 32 or less, in order to take advantage of CUDA's warp shuffle functions.
      /// \return bitset of the threads involved. If width <= 32, a vector of size 1 is returned, otherwise a vector of size width/32 is returned.
      inline vector<gpu_uint> ballot(gpu_bool value, Tuint width)
      {
	if ((width & (width-1)) != 0)
	  {
	    exwhat() << "width must be a power of 2";
	    throw EX::syntax();
	  }
          vector<gpu_uint> ret(max(width/32, 1u));
          vector<gpu_writer_base_impl*> retp(ret.size());
          for (Tuint k=0; k<retp.size(); ++k)
          {
              retp[k] = ret[k].impl.get();
          }
          check_errno(interface::goopax_E_ballot(retp.data(), width, value.impl.get()));
          return ret;
      }
      
      
    /// Returns true if b is true for all threads in the workgroup.
      /// \ingroup thread_communication
    /// \param b input value
      /// \return the resulting boolean or over the threads
    inline gpu_bool work_group_all(gpu_bool b)
    {
      gpu_bool ret;
      check_errno(interface::goopax_E_work_group_all2(ret.impl.get(), b.impl.get()));
      return ret;
    }
      /// Returns true if b is true for any thread in the workgroup.
      /// \ingroup thread_communication
      /// \param b input value
      /// \return the resulting boolean and over the threads
    inline gpu_bool work_group_any(gpu_bool b)
    {
      gpu_bool ret;
      check_errno(interface::goopax_E_work_group_any2(ret.impl.get(), b.impl.get()));
      return ret;
    }
      
      template <typename OP> typename OP::gpu_value_type scan_fallback(typename OP::gpu_value_type val, Tbool inclusive, const typename OP::gpu_value_type* nullele)
      {
          using T = typename OP::value_type;
          using gpu_T = typename gettype<T>::gpu;
          local_mem<T> mem(local_size());
          mem[local_id()] = val;
          mem.barrier();
          
          for (Tuint k=1; k<=local_size(); k*=2)
          {
              gpu_uint src =(local_id() & (~(k-1))) - 1;
              gpu_if((local_id() & k) != 0 && src >= 0)
              {
                  mem[local_id()] = OP::op((gpu_T)mem[local_id()], (gpu_T)mem[src]);
              }
              mem.barrier();
          }
          if (inclusive)
          {
              return mem[local_id()];
          }
          else
          {
              typename OP::gpu_value_type ret = *nullele;
              gpu_if(local_id() != 0)
                ret = mem[local_id()-1];
              return ret;
          }
      }

      /// reduce operation for user defined gather operator
      /// \ingroup thread_communication
      /// \param G gather operator type. See the pre-defined gather types in the file 'goopax/gather.h' as examples.
      /// \param value input value from the current thread. May be any gpu type, including goopax struct types
      /// \param width sub-group width for the gather operation. width must be a power of 2 and/or equal to local_size(). width must not exceed local_size().
      /// \return result of the gather operation
    template <typename G> typename G::gpu_value_type work_group_reduce(typename G::gpu_value_type value, Tuint width)
    {
      if ((width & (width-1)) != 0)
	{
	  if (width != local_size())
	    {
	      exwhat() << "Width must be a power of 2 and/or equal to local_size()";
	      throw EX::syntax();
	    }
	  return shuffle(scan_fallback<G>(value, true, nullptr), local_size()-1, local_size());
	}
      else
	{
	  for (Tuint subwidth=2; subwidth<=width; subwidth*=2)
	    {
	      value = G::op(value, shuffle(value, (local_id() & (subwidth-1)) ^ (subwidth/2), subwidth));
	    }
	  return value;
	}
    }

      /// reduce operation 'add'
      /// \ingroup thread_communication
      /// \param value input value
      /// \param width sub-group width for the gather operation. width must be a power of 2 and must not exceed local_size().
      /// \param return result of the gather operation
    template <typename T, typename=typename enable_if<!is_same<T, Bool>::value>::type>
    gpu_type<T> work_group_reduce_add(gpu_type<T> b, Tuint width)
    {
	if ((width & (width-1)) != 0)
	  {
	    exwhat() << "width must be a power of 2";
	    throw EX::syntax();
	  }
        // FIXME: Make this internal.
        if (interface::goopax_support_vote(0, b.impl.get()) && width == local_size())
        {
            gpu_type<T> ret;
            check_errno(interface::goopax_E_work_group_reduce_add2(ret.impl.get(), b.impl.get()));
            return ret;
        }
        else
        {
            return work_group_reduce<gather_add<T>>(b, width);
        }
    }

      /// reduce operation 'add'
      /// \ingroup thread_communication
      /// \param value input value
      /// \param width sub-group width for the gather operation. width must be a power of 2 and must not exceed local_size().
      /// \param return result of the gather operation
    template <typename gpuT>
    gpuT work_group_reduce_add(gpuT b, Tuint width)
    {
      return work_group_reduce<gather_add<typename gettype<gpuT>::cpu>>(b, width);
    }

      /// reduce operation 'min'
      /// \ingroup thread_communication
      /// \param value input value
      /// \param width sub-group width for the gather operation. width must be a power of 2 and must not exceed local_size().
      /// \param return result of the gather operation
      template <typename T>
    gpu_type<T> work_group_reduce_min(gpu_type<T> b, Tuint width)
    {
	if ((width & (width-1)) != 0)
	  {
	    exwhat() << "width must be a power of 2";
	    throw EX::syntax();
	  }
        if (interface::goopax_support_vote(1, b.impl.get()) && width == local_size())
        {
            gpu_type<T> ret;
            check_errno(interface::goopax_E_work_group_reduce_min2(ret.impl.get(), b.impl.get()));
            return ret;
        }
        else
        {
            return work_group_reduce<gather_min<T>>(b, width);
        }
    }

      /// reduce operation 'max'
      /// \ingroup thread_communication
      /// \param value input value
      /// \param width sub-group width for the gather operation. width must be a power of 2 and must not exceed local_size().
      /// \param return result of the gather operation
      template <typename T>
    gpu_type<T> work_group_reduce_max(gpu_type<T> b, Tuint width)
    {
	if ((width & (width-1)) != 0)
	  {
	    exwhat() << "width must be a power of 2";
	    throw EX::syntax();
	  }
        if (interface::goopax_support_vote(2, b.impl.get()) && width == local_size())
        {
            gpu_type<T> ret;
            check_errno(interface::goopax_E_work_group_reduce_max2(ret.impl.get(), b.impl.get()));
            return ret;
        }
        else
        {
            return work_group_reduce<gather_max<T>>(b, width);
        }
    }

      /// broadcast value from one thread to all threads in the workgroup
      /// \ingroup thread_communication
      /// \param val source value
      /// \param src_id thread number to read value from
      /// \return value that thread 'src_id' has provided
    template <typename T>
    GOOPAX_DEPRECATED // use shuffle instead.
    gpu_type<T> work_group_broadcast(gpu_type<T> val, gpu_uint src_id)
    {
      return shuffle(val, src_id, local_size());
    }

      /// Does an exclusive scan operation.
      /// \ingroup thread_communication
      /// \param val input value from the current thread
      /// \return the resulting value equals the sum of all values provided by threads 0 to local_id()-1
    template <typename T>
    gpu_type<T> work_group_scan_exclusive_add(gpu_type<T> val)
    {
        if (interface::goopax_support_vote(4, val.impl.get()))
        {
            gpu_type<T> ret;
            check_errno(interface::goopax_E_work_group_scan_exclusive_add2(ret.impl.get(), val.impl.get()));
            return ret;
        }
        else
        {
	    const gpu_type<T> nullele = 0;
            return scan_fallback<gather_add<T>>(val, false, &nullele);
        }
    }

      /// Does an inclusive scan operation.
      /// \ingroup thread_communication
      /// \param val input value from the current thread
      /// \return the resulting value equals the sum of all values provided by threads 0 to local_id()
      template <typename T>
      gpu_type<T> work_group_scan_inclusive_add(gpu_type<T> val)
      {
          if (interface::goopax_support_vote(5, val.impl.get()))
          {
              gpu_type<T> ret;
              check_errno(interface::goopax_E_work_group_scan_inclusive_add2(ret.impl.get(), val.impl.get()));
              return ret;
          }
          else
          {
              return scan_fallback<gather_add<T>>(val, true, nullptr);
          }
      }
      
      /// Does an exclusive scan operation.
      /// \ingroup thread_communication
      /// \param val input value from the current thread
      /// \return the resulting value equals the minimum of all values provided by threads 0 to local_id()-1
      template <typename T>
      gpu_type<T> work_group_scan_exclusive_min(gpu_type<T> val)
      {
          if (interface::goopax_support_vote(6, val.impl.get()))
          {
              gpu_type<T> ret;
              check_errno(interface::goopax_E_work_group_scan_exclusive_min2(ret.impl.get(), val.impl.get()));
              return ret;
          }
          else
          {
	    const gpu_type<T> nullele = (is_floating_point<T>::value ? numeric_limits<T>::infinity() : numeric_limits<T>::max());
              return scan_fallback<gather_min<T>>(val, false, &nullele);
          }
      }
      
      /// Does an inclusive scan operation.
      /// \ingroup thread_communication
      /// \param val input value from the current thread
      /// \return the resulting value equals the minimum of all values provided by threads 0 to local_id()
      template <typename T>
      gpu_type<T> work_group_scan_inclusive_min(gpu_type<T> val)
      {
          if (interface::goopax_support_vote(7, val.impl.get()))
          {
              gpu_type<T> ret;
              check_errno(interface::goopax_E_work_group_scan_inclusive_min2(ret.impl.get(), val.impl.get()));
              return ret;
          }
          else
          {
              return scan_fallback<gather_min<T>>(val, true, nullptr);
          }
      }
      
      /// Does an exclusive scan operation.
      /// \ingroup thread_communication
      /// \param val input value from the current thread
      /// \return the resulting value equals the maximum of all values provided by threads 0 to local_id()-1
      template <typename T>
      gpu_type<T> work_group_scan_exclusive_max(gpu_type<T> val)
      {
          if (interface::goopax_support_vote(8, val.impl.get()))
          {
              gpu_type<T> ret;
              check_errno(interface::goopax_E_work_group_scan_exclusive_max2(ret.impl.get(), val.impl.get()));
              return ret;
          }
          else
          {
	    const gpu_type<T> nullele = (is_floating_point<T>::value ? -numeric_limits<T>::infinity() : numeric_limits<T>::min());
              return scan_fallback<gather_max<T>>(val, false, &nullele);
          }
      }
      
      /// Does an inclusive scan operation.
      /// \ingroup thread_communication
      /// \param val input value from the current thread
      /// \return the resulting value equals the maximum of all values provided by threads 0 to local_id()
      template <typename T>
      gpu_type<T> work_group_scan_inclusive_max(gpu_type<T> val)
      {
          if (interface::goopax_support_vote(9, val.impl.get()))
          {
              gpu_type<T> ret;
              check_errno(interface::goopax_E_work_group_scan_inclusive_max2(ret.impl.get(), val.impl.get()));
              return ret;
          }
          else
          {
              return scan_fallback<gather_max<T>>(val, true, nullptr);
          }
    }
  }
  using impl::shuffle;
    using impl::ballot;
  using impl::work_group_all;
  using impl::work_group_any;
  using impl::work_group_reduce_add;
  using impl::work_group_reduce_min;
  using impl::work_group_reduce_max;
  using impl::work_group_broadcast;
  using impl::work_group_scan_exclusive_add;
  using impl::work_group_scan_inclusive_add;
  using impl::work_group_scan_exclusive_min;
  using impl::work_group_scan_inclusive_min;
  using impl::work_group_scan_exclusive_max;
  using impl::work_group_scan_inclusive_max;
}

/// \}
